﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Hive;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Clients.Hive {
  [StorableClass]
  public class RefreshableHiveExperiment : IHiveItem, IDeepCloneable, IContent, IProgressReporter {
    private JobResultPoller jobResultPoller;
    private ConcurrentJobDownloader<ItemJob> jobDownloader = new ConcurrentJobDownloader<ItemJob>(2, 2);
    private static object locker = new object();

    [Storable]
    private HiveExperiment hiveExperiment;
    public HiveExperiment HiveExperiment {
      get { return hiveExperiment; }
      set {
        if (value != hiveExperiment) {
          if (value == null)
            throw new ArgumentNullException();

          if (hiveExperiment != null) DergisterHiveExperimentEvents();
          hiveExperiment = value;
          if (hiveExperiment != null) RegisterHiveExperimentEvents();
          OnHiveExperimentChanged();
        }
      }
    }

    /** include jobs when refreshing **/
    [Storable]
    private bool includeJobs;
    public bool IncludeJobs {
      get { return includeJobs; }
      set { includeJobs = value; }
    }

    [Storable]
    private bool refreshAutomatically;
    public bool RefreshAutomatically {
      get { return refreshAutomatically; }
      set {
        lock (locker) {
          if (refreshAutomatically != value) {
            refreshAutomatically = value;
            OnRefreshAutomaticallyChanged();
          }
          if (RefreshAutomatically) {
            if (hiveExperiment.HiveJobs != null && hiveExperiment.HiveJobs.Count > 0 && (jobResultPoller == null || !jobResultPoller.IsPolling)) {
              StartResultPolling();
            }
          } else {
            StopResultPolling();
          }
        }
      }
    }

    [Storable]
    private bool isControllable = true;
    public bool IsControllable {
      get { return isControllable; }
      set {
        if (value != isControllable) {
          isControllable = value;
          OnIsControllableChanged();
        }
      }
    }

    #region Constructors and Cloning
    public RefreshableHiveExperiment() {
      this.includeJobs = true;
      this.refreshAutomatically = true;
      this.HiveExperiment = new HiveExperiment();
    }
    public RefreshableHiveExperiment(HiveExperiment hiveExperiment) {
      this.includeJobs = true;
      this.refreshAutomatically = true;
      this.HiveExperiment = hiveExperiment;
    }
    protected RefreshableHiveExperiment(RefreshableHiveExperiment original, Cloner cloner) {
      cloner.RegisterClonedObject(original, this);
      this.HiveExperiment = original.HiveExperiment;
      this.RefreshAutomatically = original.RefreshAutomatically;
      this.IncludeJobs = original.IncludeJobs;
      this.IsControllable = original.IsControllable;
    }
    public IDeepCloneable Clone(Cloner cloner) {
      return new RefreshableHiveExperiment(this, cloner);
    }
    public object Clone() {
      return this.Clone(new Cloner());
    }
    #endregion

    private void hiveExperiment_HiveJobsChanged(object sender, EventArgs e) {
      if (jobResultPoller != null && jobResultPoller.IsPolling) {
        jobResultPoller.Stop();
        DeregisterResultPollingEvents();
      }
      if (hiveExperiment.HiveJobs != null && hiveExperiment.HiveJobs.Count > 0 && hiveExperiment.GetAllHiveJobs().All(x => x.Job.Id != Guid.Empty)) {
        if (this.RefreshAutomatically)
          StartResultPolling();
      }
    }

    #region JobResultPoller Events

    public void StartResultPolling() {
      if (jobResultPoller == null) {
        jobResultPoller = new JobResultPoller(hiveExperiment.Id, /*ApplicationConstants.ResultPollingInterval*/new TimeSpan(0, 0, 5)); //TODO: find a better place for ApplicationConstants
        RegisterResultPollingEvents();
      }

      if (!jobResultPoller.IsPolling) {
        jobResultPoller.Start();
      }
    }

    public void StopResultPolling() {
      if (jobResultPoller != null && jobResultPoller.IsPolling) {
        jobResultPoller.Stop();
      }
    }

    private void RegisterResultPollingEvents() {
      jobResultPoller.ExceptionOccured += new EventHandler<EventArgs<Exception>>(jobResultPoller_ExceptionOccured);
      jobResultPoller.JobResultsReceived += new EventHandler<EventArgs<IEnumerable<LightweightJob>>>(jobResultPoller_JobResultReceived);
      jobResultPoller.IsPollingChanged += new EventHandler(jobResultPoller_IsPollingChanged);
    }
    private void DeregisterResultPollingEvents() {
      jobResultPoller.ExceptionOccured -= new EventHandler<EventArgs<Exception>>(jobResultPoller_ExceptionOccured);
      jobResultPoller.JobResultsReceived -= new EventHandler<EventArgs<IEnumerable<LightweightJob>>>(jobResultPoller_JobResultReceived);
      jobResultPoller.IsPollingChanged -= new EventHandler(jobResultPoller_IsPollingChanged);
    }
    private void jobResultPoller_IsPollingChanged(object sender, EventArgs e) {
      this.refreshAutomatically = jobResultPoller.IsPolling;
      OnRefreshAutomaticallyChanged();
    }
    private void jobResultPoller_JobResultReceived(object sender, EventArgs<IEnumerable<LightweightJob>> e) {
      foreach (LightweightJob lightweightJob in e.Value) {
        HiveJob hj = GetHiveJobById(lightweightJob.Id);
        if (hj != null) {
          DateTime lastJobDataUpdate = hj.Job.LastJobDataUpdate;
          hj.UpdateFromLightweightJob(lightweightJob);

          // lastJobDataUpdate equals DateTime.MinValue right after it was uploaded. When the first results are polled, this value is updated
          if (lastJobDataUpdate != DateTime.MinValue && lastJobDataUpdate < hj.Job.LastJobDataUpdate) {
            jobDownloader.DownloadJob(hj.Job.Id, (id, itemJob, exception) => {
              if (exception != null) {
                throw new ConcurrentJobDownloaderException("Downloading job failed.", exception);
              }

              if (itemJob == null) {
                // something bad happened to this job. bad job, BAAAD job!
              } else {
                // if the job is paused, download but don't integrate into parent optimizer (to avoid Prepare)
                if (hj.Job.State == JobState.Paused) {
                  hj.ItemJob = itemJob;
                } else {
                  if (lightweightJob.ParentJobId.HasValue) {
                    HiveJob parentHiveJob = GetHiveJobById(lightweightJob.ParentJobId.Value);
                    parentHiveJob.IntegrateChild(itemJob, hj.Job.Id);
                  } else {
                    hj.ItemJob = itemJob;
                  }
                }
              }
            });
          }
        }
      }
      GC.Collect(); // force GC, because .NET is too lazy here (deserialization takes a lot of memory)
      if (AllJobsFinished()) {
        hiveExperiment.ExecutionState = Core.ExecutionState.Stopped;
        StopResultPolling();
      }
      UpdateTotalExecutionTime();
      UpdateStats();
    }

    public HiveJob GetHiveJobById(Guid jobId) {
      foreach (HiveJob job in hiveExperiment.HiveJobs) {
        var hj = job.GetHiveJobByJobId(jobId);
        if (hj != null)
          return hj;
      }
      return null;
    }

    private void UpdateStats() {
      var jobs = hiveExperiment.GetAllHiveJobs();
      hiveExperiment.JobCount = jobs.Count();
      hiveExperiment.CalculatingCount = jobs.Count(j => j.Job.State == JobState.Calculating);
      hiveExperiment.FinishedCount = jobs.Count(j => j.Job.State == JobState.Finished);
    }

    public bool AllJobsFinished() {
      return hiveExperiment.GetAllHiveJobs().All(j => (j.Job.State == JobState.Finished
                                                   || j.Job.State == JobState.Aborted
                                                   || j.Job.State == JobState.Failed)
                                                   && j.ItemJobDownloaded);
    }

    //public bool AllJobsFinishedAndDownloaded() {
    //  return this.AllJobsFinished() && hiveExperiment.GetAllHiveJobs().All(j => j.JobItem.;
    //}

    private void jobResultPoller_ExceptionOccured(object sender, EventArgs<Exception> e) {
      //OnExceptionOccured(e.Value);
    }

    public void UpdateTotalExecutionTime() {
      hiveExperiment.ExecutionTime = TimeSpan.FromMilliseconds(hiveExperiment.GetAllHiveJobs().Sum(x => x.Job.ExecutionTime.HasValue ? x.Job.ExecutionTime.Value.TotalMilliseconds : 0));
    }
    #endregion

    private void RegisterHiveExperimentEvents() {
      hiveExperiment.HiveJobsChanged += new EventHandler(hiveExperiment_HiveJobsChanged);
      hiveExperiment.ToStringChanged += new EventHandler(hiveExperiment_ToStringChanged);
      hiveExperiment.PropertyChanged += new PropertyChangedEventHandler(hiveExperiment_PropertyChanged);
      hiveExperiment.ItemImageChanged += new EventHandler(hiveExperiment_ItemImageChanged);
      hiveExperiment.ModifiedChanged += new EventHandler(hiveExperiment_ModifiedChanged);
      hiveExperiment.IsProgressingChanged += new EventHandler(hiveExperiment_IsProgressingChanged);
      hiveExperiment.Loaded += new EventHandler(hiveExperiment_Loaded);
    }

    private void DergisterHiveExperimentEvents() {
      hiveExperiment.HiveJobsChanged -= new EventHandler(hiveExperiment_HiveJobsChanged);
      hiveExperiment.ToStringChanged -= new EventHandler(hiveExperiment_ToStringChanged);
      hiveExperiment.PropertyChanged -= new PropertyChangedEventHandler(hiveExperiment_PropertyChanged);
      hiveExperiment.ItemImageChanged -= new EventHandler(hiveExperiment_ItemImageChanged);
      hiveExperiment.ModifiedChanged -= new EventHandler(hiveExperiment_ModifiedChanged);
      hiveExperiment.IsProgressingChanged -= new EventHandler(hiveExperiment_IsProgressingChanged);
      hiveExperiment.Loaded -= new EventHandler(hiveExperiment_Loaded);
    }

    private void hiveExperiment_Loaded(object sender, EventArgs e) {
      this.UpdateTotalExecutionTime();

      if (hiveExperiment.ExecutionState != ExecutionState.Stopped) {
        this.RefreshAutomatically = true;
      }
    }

    #region Events
    public event EventHandler RefreshAutomaticallyChanged;
    private void OnRefreshAutomaticallyChanged() {
      var handler = RefreshAutomaticallyChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler HiveExperimentChanged;
    private void OnHiveExperimentChanged() {
      var handler = HiveExperimentChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler ModifiedChanged;
    private void hiveExperiment_ModifiedChanged(object sender, EventArgs e) {
      var handler = ModifiedChanged;
      if (handler != null) handler(sender, e);
    }

    public event EventHandler ItemImageChanged;
    private void hiveExperiment_ItemImageChanged(object sender, EventArgs e) {
      var handler = ItemImageChanged;
      if (handler != null) handler(this, e);
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void hiveExperiment_PropertyChanged(object sender, PropertyChangedEventArgs e) {
      var handler = PropertyChanged;
      if (handler != null) handler(sender, e);
    }

    public event EventHandler ToStringChanged;
    private void hiveExperiment_ToStringChanged(object sender, EventArgs e) {
      var handler = ToStringChanged;
      if (handler != null) handler(this, e);
    }

    public event EventHandler IsProgressingChanged;
    private void hiveExperiment_IsProgressingChanged(object sender, EventArgs e) {
      var handler = IsProgressingChanged;
      if (handler != null) handler(sender, e);
    }

    public event EventHandler IsControllableChanged;
    private void OnIsControllableChanged() {
      var handler = IsControllableChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    #endregion

    public Guid Id {
      get { return hiveExperiment.Id; }
      set { hiveExperiment.Id = value; }
    }
    public bool Modified {
      get { return hiveExperiment.Modified; }
    }
    public void Store() {
      hiveExperiment.Store();
    }
    public string ItemDescription {
      get { return hiveExperiment.ItemDescription; }
    }
    public Image ItemImage {
      get { return hiveExperiment.ItemImage; }
    }
    public string ItemName {
      get { return hiveExperiment.ItemName; }
    }
    public Version ItemVersion {
      get { return hiveExperiment.ItemVersion; }
    }


    #region IProgressReporter Members
    public IProgress Progress {
      get { return HiveExperiment.Progress; }
    }

    public bool IsProgressing {
      get { return HiveExperiment.IsProgressing; }
    }
    #endregion

    public override string ToString() {
      return HiveExperiment.ToString();
    }
  }
}
